#!/usr/bin/env python

"""
OpenID redirection classes, sending unauthenticated users to the OpenID
initiation page.

Copyright (C) 2004, 2005, 2006, 2007 Paul Boddie <paul@boddie.org.uk>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from WebStack.Resources.LoginRedirect import LoginRedirectResource
from WebStack.Helpers.Auth import Verifier, check_openid_signature
import WebStack.Generic
import datetime
import urllib

class OpenIDRedirectResource(LoginRedirectResource):

    "A resource redirecting to an OpenID initiation page."

    openid_ns = "http://specs.openid.net/auth/2.0"

    def respond(self, trans):

        "Respond using the given transaction 'trans'."

        fields = trans.get_fields(self.path_encoding)

        # If requested, attempt to verify OpenID assertions.
        # http://openid.net/specs/openid-authentication-2_0-12.html#rfc.section.11

        if fields.get("openid.mode", [None])[0] == "id_res":

            # The additional condition could be used to insist on OpenID 2.0
            # conformance:
            # fields.get("openid.ns", [None])[0] == self.openid_ns

            if self.authenticator.authenticate(trans, verify=1):
                trans.redirect(fields["openid.return_to"][0])

        # Otherwise, handle the usual parameters and request details.

        LoginRedirectResource.respond(self, trans)

class OpenIDRedirectAuthenticator(Verifier):

    """
    An authenticator which verifies the credentials provided in a special login
    cookie, accepting OpenID assertions if necessary.
    """

    encoding = "utf-8"
    openid_ns = "http://specs.openid.net/auth/2.0"
    replay_limit = datetime.timedelta(0, 10) # 10s

    def __init__(self, secret_key, app_url, associations=None, replay_limit=None,
        cookie_name=None, urlencoding=None):

        """
        Initialise the authenticator with a 'secret_key', 'app_url' and optional
        'associations', 'replay_limit', 'cookie_name' and 'urlencoding'.
        """

        Verifier.__init__(self, secret_key, cookie_name)

        self.app_url = app_url
        self.associations = associations or {}
        self.replay_limit = replay_limit or self.replay_limit
        self.urlencoding = urlencoding or self.encoding

    def authenticate(self, trans, verify=0):

        """
        Authenticate the originator of 'trans', updating the object if
        successful and returning a true value if successful, or a false value
        otherwise.

        If the optional 'verify' parameter is specified as a true value, perform
        verification on any 
        """

        # If requested, attempt to verify OpenID assertions.

        if verify:
            fields = trans.get_fields(self.urlencoding)

            # NOTE: Could expose all errors.

            try:
                # Test the details of the assertion.

                if self.test_url(fields) and \
                    self.test_signature(fields) and \
                    self.test_replay(fields):

                    self.set_token(trans, fields["openid.identity"][0])
                    return 1

            # Incomplete assertion.

            except (KeyError, ValueError):
                raise

            # Assertion failed or was incomplete.

            return 0

        # Otherwise, try to authenticate with an application cookie.

        else:
            valid = Verifier.authenticate(self, trans)

            # Update the transaction with the user details.

            if valid:
                username, token = self.get_username_and_token(trans)
                trans.set_user(username)
            return valid

    def test_url(self, fields):

        """
        See:
        http://openid.net/specs/openid-authentication-2_0-12.html#rfc.section.11.1
        """

        # NOTE: Currently, this is not strict enough.

        return fields["openid.return_to"][0].startswith(self.app_url)

    def test_signature(self, fields):

        """
        See:
        http://openid.net/specs/openid-authentication-2_0-12.html#rfc.section.11.4
        http://openid.net/specs/openid-authentication-2_0-12.html#rfc.section.6
        """

        handle = fields.get("openid.assoc_handle", [None])[0]

        # With an association handle, look up the appropriate secret key and
        # verify the signed items.

        if handle is not None:

            # Where an association exists, use the known secret key.

            if self.associations.has_key(handle):
                return check_openid_signature(fields, self.associations[handle])

            # Without an association, request verification of the signed items
            # from the OpenID provider.

            else:
                return self.test_signature_direct(fields)

        # Without a handle, no signature verification can occur.

        return 0

    def test_signature_direct(self, fields):

        """
        See:
        http://openid.net/specs/openid-authentication-2_0-12.html#rfc.section.11.4.2
        """

        # Make a POST request using the "openid." fields.

        d = {}
        for name, values in fields.items():
            if name.startswith("openid.") and name != "openid.mode":
                d[name] = values[0]
        d["openid.mode"] = "check_authentication"
        data = urllib.urlencode(d)

        # Send a POST request to the OpenID provider, reading the response and
        # testing for certain fields and values.

        f = urllib.urlopen(fields["openid.op_endpoint"][0], data)
        try:
            items = []
            for line in f.readlines():
                if line[-1] == "\n":
                    line = line[:-1]
                parts = line.split(":")
                items.append((parts[0], ":".join(parts[1:])))
            fields = dict(items)
            return fields["ns"] == self.openid_ns and fields["is_valid"] == "true"
        finally:
            f.close()

    def test_replay(self, fields):

        """
        See:
        http://openid.net/specs/openid-authentication-2_0-12.html#rfc.section.11.3
        """

        timestamp = fields["openid.response_nonce"][0]
        # YYYY-MM-DDTHH:MM:SSZ...
        year, month, day, hour, minute, second, unique = \
            int(timestamp[0:4]), int(timestamp[5:7]), int(timestamp[8:10]), \
            int(timestamp[11:13]), int(timestamp[14:16]), int(timestamp[17:19]), \
            timestamp[20:]
        dt = datetime.datetime(year, month, day, hour, minute, second)
        return -self.replay_limit < (datetime.datetime.utcnow() - dt) < self.replay_limit

    def set_token(self, trans, username):

        "Set an authentication token in 'trans' with the given 'username'."

        Verifier.set_token(self, trans, username)

        # Update the transaction with the user details.

        trans.set_user(username)

# vim: tabstop=4 expandtab shiftwidth=4
